/*
 *      c2html.c
 *      C Quelltext nach HTML konvertieren
 *      1.00, Klingebiel, Tue Dec 23 08:04:07 1997
 *      1.01, Klingebiel, Thu Nov 18 09:48:02 1999
 *
 *      Aufruf: c2html [-a][-c][-h][-l][-n][-t] [infile.c] [outfile.html]
 */

static const char *sccs = "@(#) c2html.c 1.01 (klin) Dec 23 1999";
static const char *prgname = "c2html";
static const char *prgvers = "1.01";

/*
 *      Headerdateien        
 */
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*
 *      Farbdefinitionen
 */

#include "c2html.h"

static char *bgcol = BGCOL;             /* Hintergrund          */
static char *tdcol = TDCOL;             /* Tabellenhintergrund  */
static const char *txcol = TXCOL;       /* Text                 */
static const char *rwcol = RWCOL;       /* Schluesselworte      */
static const char *sccol = SCCOL;       /* Strings/Zeichen      */
static const char *cmcol = CMCOL;       /* Kommentare           */
static const char *lncol = LNCOL;       /* Zeilennummern        */

/* 
 *      Reservierte Worte in C und CPP 
 *      und die Groessen der Tabellen
 */
 
static const char *ckeytab[] = {
  "auto",       "break",        "case",
  "char",       "const",        "continue",
  "default",    "do",           "double",
  "else",       "enum",         "extern",
  "float",      "for",          "goto",
  "if",         "int",          "long",
  "register",   "return",       "short",
  "signed",     "sizeof",       "static",
  "struct",     "switch",       "typedef",
  "union",      "unsigned",     "void",
  "volatile",   "while",
};

static const int nckeys = (sizeof(ckeytab) / sizeof(char *));

static const char *pkeytab[] = {
  "define",     "defined",      "elif",         "else",
  "endif",      "error",        "if",
  "ifdef",      "ifndef",       "include",
  "line",       "pragma",       "undef",
};

static const int npkeys = (sizeof(pkeytab) / sizeof(char *));

/*
 *      Symbole fuer Zeichentypen wie sie
 *      von getword() geliefert werden
 */
 
#define LETTER  0x0100          /* Zeichen/Wort         */
#define DIGIT   0x0200          /* Ziffer               */
#define CSTART  0x0300          /* Kommentaranfang      */
#define CEND    0x0400          /* Kommentarende        */
#define MASK    0x0500          /* Maskierung           */
#define BSTART  0x0600          /* Blockanfang          */
#define BEND    0x0700          /* Blockende            */
#define DSTART  0x0800          /* Stringanfang         */
#define DEND    0x0900          /* Stringende           */
#define SSTART  0x0a00          /* Zeichenanfang        */
#define SEND    0x0b00          /* Zeichenende          */

/*
 *      Weitere Konstanten
 */
 
#define MAXWORD 512             /* Maximale Wortlaenge  */

/* 
 *      Zustandsflags
 */
 
static int comment = 0;         /* Kommentarflag        */
static int squote  = 0;         /* Anfuehrungszeichen   */
static int dquote  = 0;         /* Anfuehrungszeichen   */
static int hash    = 0;         /* Hash-Zeichen         */
static int newline = 0;         /* Neue Zeile           */

/* 
 *      Prototypen 
 */

static void c2html(int, int);
static int  binary(char *, const char *[], const int);
static int  getword(char *, int);
static int  type(int);
static void putheader(char *, int, int, int);
static void puttrailer(int);
static int  iseof(void);

/* 
 *      Main 
 */
 
void main(int argc, char *argv[])
{
  int cflag = 0;        /* Colorflag            */
  int lflag = 0;        /* Linkflag             */
  int nflag = 0;        /* Zeilennummerflag     */
  int tflag = 0;        /* Tabellenflag         */

  /* Optionen pruefen */
  while (argc > 1 && argv[1][0] == '-') {
    /* Zeilennummer zaehlen */
    if (strcmp(argv[1], "-n") == 0) {
      nflag = 1;
      --argc;
      ++argv;
    }
    /* Farben in Ausgabe nutzen */
    else if (strcmp(argv[1], "-c") == 0) {
      cflag = 1;
      --argc;
      ++argv;
    }
    /* Tabelle in Ausgabe nutzen */
    else if (strcmp(argv[1], "-t") == 0) {
      tflag = 1;
      --argc;
      ++argv;
    }
    /* Link auf Quelle setzen */
    else if (strcmp(argv[1], "-l") == 0) {
      lflag = 1;
      --argc;
      ++argv;
    }
    /* Alle Optionen nutzen */
    else if (strcmp(argv[1], "-a") == 0) {
      cflag = lflag = nflag = tflag = 1;
      --argc;
      ++argv;
    }
    /* Ende der Optionen */
    else if (strcmp(argv[1], "--") == 0) {
      --argc;
      ++argv;
      break;
    }
    /* Hilfe */
    else if (strcmp(argv[1], "-h") == 0) {
      fprintf(stderr, "Aufruf: %s [-a][-c][-h][-l][-n][-t] [infile.c] [outfile.html]\n", prgname);
      fprintf(stderr, "\t-a\tAlle Optionen (clnt) einschalten\n");
      fprintf(stderr, "\t-c\tSchluesselworte und Kommentare in Farbe ausgeben\n");
      fprintf(stderr, "\t-h\tHilfe zu Aufruf und Optionen ausgeben\n");
      fprintf(stderr, "\t-l\tLink auf Quelldatei setzen\n");
      fprintf(stderr, "\t-n\tZeilennummern ausgeben\n");
      fprintf(stderr, "\t-t\tTabelle nutzen\n");
      exit(0);
    }
    /* Falsche Option */
    else {
      fprintf(stderr, "Aufruf: %s [-a][-c][-h][-l][-n][-t] [infile.c] [outfile.html]\n", prgname);
      exit(1);
    }
  }
  
  /* Parameter pruefen */
  if (argc > 1 && freopen(argv[1], "r", stdin) == NULL) {
    fprintf(stderr, "%s: Kann Datei %s nicht oeffnen\n", prgname, argv[1]);
    /* Besser: perror(argv[1]); */
    exit(1);
  }
  if (argc > 2 && freopen(argv[2], "w", stdout) == NULL) {
    fprintf(stderr, "%s: Kann Datei %s nicht oeffnen\n", prgname, argv[2]);
    /* Besser: perror(argv[2]); */
    exit(1);
  }
  if (tflag) {
    bgcol = TDCOL;
    tdcol = BGCOL;
  }

  /* HTML-Header ausgeben */
  putheader(argc > 1 ? argv[1] : NULL, cflag, lflag, tflag);

  /* Aktion: Lesen, Formatieren, Schreiben */
  c2html(cflag, nflag);

  /* HTML-Trailer ausgeben */
  puttrailer(tflag);

  exit(0);

} /* main() */

/* 
 *      C wortweise lesen und nach HTML konvertieren 
 */

static void c2html(int cflag, int nflag)
{    
  char word[MAXWORD];   /* Eingabewort          */
  int t;                /* Zeichen/Symbol       */
  int nline = 1;        /* Zeilennummer         */
	    
  /* Ggfs. Zeilennummer ausgeben */
  if (nflag && !iseof()) {
    if (cflag)
      fprintf(stdout, "<FONT COLOR=%s><I>%5d   </I></FONT>", lncol, nline);
    else
      fprintf(stdout, "<I>%5d   </I>", nline);
  }
    
  /* Bis EOF wort-/zeichenweise lesen */
  while ((t = getword(word, MAXWORD)) != EOF) {
    switch (t) {
      case LETTER:      /* Wort */
	/* Kommentar oder String */
	if (comment || dquote || squote) {
	  fprintf(stdout, "%s", word);
	}
	/* CPP-Schluesselwort */
	else if (hash && binary(word, pkeytab, npkeys) >= 0) {
	  if (cflag)
	    fprintf(stdout, "<FONT COLOR=%s><B>%s</B></FONT>", rwcol, word);
	  else
	    fprintf(stdout, "<B>%s</B>", word);
	}
	/* C-Schluesselwort */
	else if (!hash && binary(word, ckeytab, nckeys) >= 0) {
	  if (cflag)
	    fprintf(stdout, "<FONT COLOR=%s><B>%s</B></FONT>", rwcol, word);
	  else
	    fprintf(stdout, "<B>%s</B>", word);
	}
	/* Sonstiges Wort */
	else {
	  fprintf(stdout, "%s", word);
	}
	break;
      case CSTART:      /* Kommentaranfang */
	if (cflag)
	  fprintf(stdout, "<FONT COLOR=%s><I>%s", cmcol, word);
	else
	  fprintf(stdout, "<I>%s", word);
	break;
      case CEND:        /* Kommentarende */
	if (cflag)
	  fprintf(stdout, "%s</I></FONT>", word);
	else
	  fprintf(stdout, "%s</I>", word);
	break;
      case BSTART:      /* Blockanfang */
	fprintf(stdout, "<B>%s</B>", word);
	break;
      case BEND:        /* Blockende */
	fprintf(stdout, "<B>%s</B>", word);
	break;
      case DSTART:      /* Stringanfang */
	if (cflag)
	  fprintf(stdout, "<FONT COLOR=%s>&quot;", sccol);
	else
	  fprintf(stdout, "&quot;");
	break;
      case DEND:        /* Stringende */
	if (cflag)
	  fprintf(stdout, "&quot;</FONT>");
	else
	  fprintf(stdout, "&quot;");
	break;
      case SSTART:      /* Zeichenanfang */
	if (cflag)
	  fprintf(stdout, "<FONT COLOR=%s>\'", sccol);
	else
	  fprintf(stdout, "\'");
	break;
      case SEND:        /* Zeichenende */
	if (cflag)
	  fprintf(stdout, "\'</FONT>");
	else
	  fprintf(stdout, "\'");
	break;
      case MASK:        /* Maskierung */
	fprintf(stdout, "%s", word);
	break;
      case '\"':        /* HTML-Sonderzeichen -> Entity */
	fprintf(stdout, "&quot;");
	break;
      case '&':         /* HTML-Sonderzeichen -> Entity */
	fprintf(stdout, "&amp;");
	break;
      case '<':         /* HTML-Sonderzeichen -> Entity */
	fprintf(stdout, "&lt;");
	break;
      case '>':         /* HTML-Sonderzeichen -> Entity */
	fprintf(stdout, "&gt;");
	break;
      case '\n':        /* Zeilenende */
	if (nflag && !iseof()) {
	  ++nline;
	  if (comment) {
	    if (cflag) {
	      fprintf(stdout, "</I></FONT>");
	      fprintf(stdout, "\n<FONT COLOR=%s><I>%5d   </I></FONT>", lncol, nline);
	      fprintf(stdout, "<FONT COLOR=%s><I>", cmcol);
	    }
	    else
	      fprintf(stdout, "</I>\n<I>%5d   </I><I>", nline);
	  }
	  else
	    fprintf(stdout, "\n<FONT COLOR=%s><I>%5d   </I></FONT>", lncol, nline);
	  break;
	}        
	/* Sonst: NL ausgeben wie sonstige Zeichen */
      default:          /* Sonstiges Zeichen */
	fprintf(stdout, "%c", t);
	break;
    }
  }

} /* c2html() */

/* 
 *      HTML-Header ausgeben 
 */
 
static void putheader(char *name, int cflag, int lflag, int tflag)
{
  fprintf(stdout, "<HTML>\n");
  fprintf(stdout, "<HEAD>\n");
  if (name)
    fprintf(stdout, "<TITLE>Datei: %s</TITLE>\n", name);
  else
    fprintf(stdout, "<TITLE>C Quelltext</TITLE>\n");
  if (cflag)
    fprintf(stdout, "<BODY BGCOLOR=\"%s\" TEXT=%s>\n", bgcol, txcol);
  else
    fprintf(stdout, "<BODY>\n");
  if (tflag) {
    fprintf(stdout, "<TABLE BORDER=\"0\" CELLSPACING=\"0\" CELLPADDING=\"6\" WIDTH=\"100%%\" HEIGHT=\"100%%\">");
    fprintf(stdout, "<TR>\n<TD BGCOLOR=\"%s\" WIDTH=\"10%%\">&nbsp;</TD>\n", bgcol);
    fprintf(stdout, "<TD WIDTH=\"80%%\" VALIGN=\"top\" BGCOLOR=\"%s\">\n", tdcol);
  }
  if (name) {
    if (lflag)
      fprintf(stdout, "<H2>Datei: <A HREF=\"%s\">%s</A></H2>\n<HR>\n", name, name);
    else
      fprintf(stdout, "<H2>Datei: %s</H2>\n<HR>\n", name);
  }
  fprintf(stdout, "<P>\n");
  fprintf(stdout, "<PRE>\n");

} /* putheader() */

/* 
 *      HTML-Trailer ausgeben 
 */
 
static void puttrailer(int tflag)
{
  fprintf(stdout, "</PRE>\n");
  fprintf(stdout, "<P>\n");
  fprintf(stdout, "<HR>\n");
  fprintf(stdout, "<I>Erzeugt von %s %s</I>\n", prgname, prgvers);
  if (tflag)
    fprintf(stdout, "</TD>\n<TD BGCOLOR=\"%s\" WIDTH=\"10%%\">&nbsp;</TD>\n</TR>\n</TABLE>\n", bgcol);
  fprintf(stdout, "</BODY>\n");
  fprintf(stdout, "</HTML>\n");

} /* puttrailer() */

/* 
 *      Wort word in lexikalisch sortierter Tabelle tab finden 
 */
 
static int binary(char *word, const char *tab[], const int n)
{
  int low, high, mid, cond;

  low = 0;
  high = n - 1;
  while (low <= high) {
    mid = (low + high) / 2;
    cond = strcmp(word, tab[mid]);
    if (cond < 0)
      high = mid - 1;
    else if (cond > 0)
      low = mid + 1;
    else
      return(mid);
  }
  return(-1);

} /* binary() */

/* 
 *      Naechstes Wort aus der Eingabe holen
 */
 
static int getword(char *w, int lim)
{
  int c, cc, t;

  c = getc(stdin);
  /* Kein Wort: Einzelzeichen! */
  if (type(c) != LETTER) {
    /* Newlineflag/Hashflag pruefen und ggfs. ruecksetzen */
    if (c == '\n') {
      newline = 1;
      hash = 0;
    }
    else if (newline && !comment && !dquote && !squote && c == '#') {
      hash = 1;
    }
    else if (newline) {
      newline = 0;
    }
    /* Maskierung */
    if (c == '\\' && !comment) {
      cc = getc(stdin);
      *w++ = c;
      *w++ = cc;
      if (type(cc) == DIGIT) {
	while (--lim > 2) {
	  c = getc(stdin);
	  t = type(c);
	  if (t != DIGIT) {
	    ungetc(c, stdin);
	    break;
	  }
	  *w++ = c;
	}
      }
      *w = '\0';
      return(MASK);
    }
    /* Einfache Anfuehrungszeichen */
    else if (c == '\'' && !comment && !dquote) {
      if (squote) {
	squote = 0;
	return(SEND);
      }
      else {
	squote = 1;
	return(SSTART);
      }
    }
    /* Doppelte Anfuehrungszeichen */
    else if (c == '\"' && !comment && !squote) {
      if (dquote) {
	dquote = 0;
	return(DEND);
      }
      else {
	dquote = 1;
	return(DSTART);
      }
    }
    /* Blockanfang */
    else if (c == '{' && !comment && !dquote && !squote) {
      *w++ = c;
      *w = '\0';
      return(BSTART);
    }
    /* Blockende */
    else if (c == '}' && !comment && !dquote && !squote) {
      *w++ = c;
      *w = '\0';
      return(BEND);
    }
    /* Kommentaranfang ? */
    else if (c == '/' && !comment && !dquote && !squote) {
	if ((cc = getc(stdin)) == '*') {
	  comment = 1;
	  *w++ = c;
	  *w++ = cc;
	  *w = '\0';
	  return(CSTART);
	}
	ungetc(cc, stdin);
    }
    /* Kommentarende ? */
    else if (c == '*' && comment && !dquote && !squote) {
      if ((cc = getc(stdin)) == '/') {
	comment = 0;
	*w++ = c;
	*w++ = cc;
	*w = '\0';
	return(CEND);
      }
      ungetc(cc, stdin);
    }
    *w++ = c;
    *w = '\0';
    return(c);
  }
  /* Wort lesen */
  *w++ = c;
  while (--lim > 0) {
    c = getc(stdin);
    t = type(c);
    if (t != LETTER && t != DIGIT) {
      ungetc(c, stdin);
      break;
    }
    *w++ = c;
  }
  *w = '\0';
  return(LETTER);

} /* getword() */

/* 
 *      Typ eines ASCII Zeichens liefern 
 */
 
static int type(int c)
{
  if (c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z' || c == '_')
    return(LETTER);
  else if (c >= '0' && c <= '9')
    return(DIGIT);
  else
    return(c);

} /* type() */

/* 
 *      Pruefen, ob naechstes Zeichen EOF 
 */
 
static int iseof(void)
{
  int c;
  
  if ((c = getc(stdin)) == EOF)
    return(1);
  ungetc(c, stdin);
  return(0);
  
} /* iseof() */
