/*
 *      ipaddr.c
 *      Anzahl Subnetzbits einlesen,
 *      daraus Netzmaske und Broadcast bilden,
 *      dann IP-Adressen einlesen und pruefen
 *      klin, Thu Nov 24 19:51:52 1999
 */

#include <stdio.h>

/* Typ fuer 32-Bit IP-Addresse  */
#define IPADDR unsigned long

/* 32-Bit Maske fuer das untere Byte    */
#define IPMASK 0x000000ff

/* Ausgabeflag dezimal, sedezimal, oktal, binaer */
#define PUTDEC  0
#define PUTHEX  1
#define PUTOCT  2
#define PUTBIN  3

/* Prototypen */
IPADDR get_ipaddr(void);
void put_ipaddr(IPADDR, int);
void put_bits(unsigned char);

void main(void)
{
  IPADDR mask;  /* Netzmaske            */
  IPADDR cast;  /* Broadcast-Adresse    */
  IPADDR subn;  /* Subnetzadresse       */
  IPADDR ia;    /* Allg. IP-Adresse     */
  int nbits;    /* Bits des Subnets     */

  /* Subnet-Bits lesen und Netzmaske bilden */
  printf("Subnet-Bits eingeben: ");
  scanf("%d", &nbits);
  mask = ~0 << nbits;

  /* IP-Adresse lesen und Broadcast bilden */
  printf("IP-Adresse eingeben:  ");
  cast = get_ipaddr() | ~mask;

  /* Subnetzaddresse ermitteln */
  subn = mask & cast;
  
  /* Konfiguration anzeigen */
  printf("\nKonfiguration\n=============\n");
  printf("\nNetzmaske: ");
  put_ipaddr(mask, PUTDEC);
  printf("\nBroadcast: ");
  put_ipaddr(cast, PUTDEC);
  printf("\nSubnetz:   ");
  put_ipaddr(subn, PUTDEC);
  printf("\n\n");

  /* IP-Addressen lesen und pruefen */
  do {
    printf("IP-Addresse eingeben: ");
    ia = get_ipaddr();
    put_ipaddr(ia, PUTDEC);
    printf(" - ");
    if ((ia & mask) == subn) {
      printf("im Subnetz ");
      put_ipaddr(subn, PUTDEC);
    }
    else {
      printf("nicht im Subnetz ");
      put_ipaddr(subn, PUTDEC);
    }
    printf("\n");
  } while (ia != 0);

} /* main() */

/* IP-Adresse xxx.xxx.xxx.xxx einlesen */
IPADDR get_ipaddr(void)
{
  IPADDR ra;
  unsigned int d1, d2, d3, d4;

  /* IP-Addresse in Dezimalform einlesen */
  scanf("%d.%d.%d.%d", &d4, &d3, &d2, &d1);

  /* IP-Adresse in Variable kopieren ... */
  ra = 0;
  ra = d4 << 24 | d3 << 16 | d2 << 8 | d1;
  /* ... und zurueckgeben */

  return(ra);

} /* get_ipaddr() */

/* IP-Adresse als xxx.xxx.xxx.xxx ausgeben */
void put_ipaddr(IPADDR a, int flag)
{
  int d4, d3, d2, d1;

  /* IP-Adresse in 4 Bytes aufbrechen ... */
  d4 = a >> 24 & IPMASK;
  d3 = a >> 16 & IPMASK;
  d2 = a >>  8 & IPMASK;
  d1 = a       & IPMASK;

  /* ... und ausgeben */
  switch (flag) {
    case PUTHEX:        /* sedezimal */
      printf("%02X.%02X.%02X.%02X", d4, d3, d2, d1);
      break;
    case PUTOCT:        /* oktal */
      printf("%02o.%02o.%02o.%02o", d4, d3, d2, d1);
      break;
    case PUTBIN:        /* binaer */
      put_bits(d4);
      putchar('.');
      put_bits(d3);
      putchar('.');
      put_bits(d2);
      putchar('.');
      put_bits(d1);
      break;
    default:            /* dezimal */
      printf("%d.%d.%d.%d", d4, d3, d2, d1);
      break;
  }

} /* put_ipaddr() */

/* Byte bitweise ausgeben */
void put_bits(unsigned char c)
{
  unsigned char m;      /* Maske   */

  /* Maske und Zaehler initialisieren */
  m = 1 << 7;
  /* Schleife ueber alle Bitpositionen */
  while (m > 0) {
    /* 1, wenn Bit gesetzt, sonst 0 */
    putchar(c & m ? '1' : '0');
    /* Maske verschieben */
    m >>= 1;
  }

} /* put_bits() */
