/*
 *      dlist.c
 *      Strings bis EOF einlesen, dynamisch in doppelt verketteter
 *      Liste mit Zeilennummer und Stringlaenge speichern,
 *      Liste mit Komponenten vom Enden her auf stdout ausgeben
 *      klin, Thu Jan  6 08:44:34 2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NBUF 1024       /* Groesse des Einlesepuffers   */

typedef struct _slist { /* Strukt fuer Stringliste      */
  char *str;            /*   String                     */
  size_t len;           /*   Stringlaenge               */
  int lno;              /*   Zeilennummer               */
  struct _slist *prv;   /*   Voriges Listenelement      */
  struct _slist *nxt;   /*   Naechstes Listenelement    */
} slist_t;

void main(void)
{
  char str[NBUF];       /* Einlesepuffer                */
  slist_t *last = NULL; /* Letztes Listenelement        */
  slist_t *curr = NULL; /* Aktuelles Listenelement      */
  int lno = 0;          /* Zeilennummer                 */
  size_t len;           /* Stringlaenge                 */

  /* 
   *    Strings bis EOF von stdin lesen 
   *    und in dynamische liste eintragen
   */
   
  while (fgets(str, sizeof(str), stdin) != NULL) {
  
    /*
     *  1.) String, Stringlaenge, Zeilennummer 
     */
       
    /* Zeilennummer hochzaehlen */
    ++lno;
    /* Stringlaenge feststellen, NL ('\n') entfernen */
    len = strlen(str);
    if (str[len - 1] == '\n')
      str[--len] = '\0';

    /* 
     *  2.) Neues Listenelement allozieren 
     */
     
    curr = (slist_t *) calloc(1, sizeof(slist_t));
    /* Fehler: Kein Speicherplatz */
    if (curr == NULL) {
      fprintf(stderr, "u7-1: Fehler: Kein Speicherplatz!\n");
      exit(1);
    }   
    /* Platz fuer String allozieren */
    curr->str = (char *) calloc(len + 1, sizeof(char));
    /* Fehler: Kein Speicherplatz */
    if (curr->str  == NULL) {
      fprintf(stderr, "u7-1: Fehler: Kein Speicherplatz!\n");
      exit(1);
    }
    
    /*
     *  3.) Komponenten des neuen Elements besetzen
     */
     
    /* String in Listekomponente kopieren */
    (void) strcpy(curr->str, str);
    /* Stringlaenge und Zeilennummer eintragen */
    curr->len = len;
    curr->lno = lno;
    /* Zeiger auf naechstes Element initialisieren */
    curr->nxt = NULL;
    
    /* 
     *  4.) Liste verknuepfen 
     */
     
    /* Zeiger auf voriges Element: NULL, wenn 1. Eintrag */
    curr->prv = last;
    /* Wenn vorhanden, Vorgaenger mit Liste verknuepfen */
    if (last != NULL) 
      last->nxt = curr;

    /* Listenelement fuer die naechste Runde merken! */
    last = curr;    
    
  } /* while */
  
  /* 
   *    Liste rueckwaerts mit Zeilennummer und Stringlaenge ausgeben 
   */
   
  /* Ende der Liste suchen */
  for (curr = last; curr->nxt != NULL; curr = curr->nxt)
    ;
  /* Ausgeben und rueckwaerts laufen */
  for ( ; curr != NULL; curr = curr->prv) 
    printf("%4d: %s (%d)\n", curr->lno, curr->str, curr->len);

  /*
   *    Programm terminieren
   */
   
  exit(0);
      
} /* main() */
